from pgkit.application.models import Postgres
from pgkit.application.utils import *
from jinja2 import Template
from pathlib import Path
from time import sleep


class Replica(Postgres):
    def __init__(self, master, port, delay):
        super().__init__(name=master.name,
                         host='localhost',
                         port=port,
                         dbname='postgres',
                         version=master.version,
                         username='postgres',
                         password=None,
                         slot=None)
        self.master = master
        self.delay = delay
        self.db_location = '/var/lib/postgresql/{}/{}'.format(self.version, self.name)
        self.wal_location = '/var/lib/postgresql/wals/{}/{}'.format(self.version, self.name)

    def start_backup(self):
        self.create_cluster()
        self.stop()
        self.remove_db_directory()
        self.setup_wal_receive_service()
        self.master.force_switch_wal()
        sleep(10)  # time to receive first wal segments
        self.create_base_backup()
        self.configure_recovery_file()
        self.start()

    def create_cluster(self):
        execute_sync('pg_createcluster {} {}'.format(self.version, self.name))

    def stop(self):
        execute_sync('pg_ctlcluster {} {} stop'.format(self.version, self.name))

    def start(self):
        execute_sync('pg_ctlcluster {} {} start'.format(self.version, self.name))

    def remove_db_directory(self):
        execute_sync('rm -rf {}'.format(self.db_location))

    def _stop_old_wal_receive_service(self):
        try:
            stop_service(f'receivewal-{self.version}-{self.name}')
        except:
            pass

    def setup_wal_receive_service(self):
        self._stop_old_wal_receive_service()

        template_path = Path(__file__).parent / "../templates/wal-receive-service.service"
        template = Template(read_file(template_path))

        print('Create wal folder')
        create_directory(self.wal_location)

        print('Create wal-receive service')
        service_file = template.render(
            name=self.name,
            version=self.version,
            host=self.master.host,
            port=self.master.port,
            username=self.master.username,
            password=self.master.password,
            slot=self.master.slot,
            destination=self.wal_location
        )
        write_file(f'/etc/systemd/system/receivewal-{self.version}-{self.name}.service', service_file)

        print('Run wal-receive service')
        restart_service(f'receivewal-{self.version}-{self.name}', reload_systemctl=True)

    def create_base_backup(self):
        print('Take basebackup')
        command = f'/usr/lib/postgresql/{self.version}/bin/pg_basebackup' \
                  f' -h {self.master.host}' \
                  f' -p {self.master.port}' \
                  f' -D {self.db_location}' \
                  f' -U {self.master.username}' \
                  f' -v --checkpoint=fast --progress'
        execute_sync(command, env=[('PGPASSWORD', self.master.password)])

        print('change owner to postgres')
        chown(self.db_location, 'postgres')

    def configure_recovery_file(self):
        template_path = Path(__file__).parent / "../templates/{}-recovery.conf".format(self.version)
        template = Template(read_file(template_path))

        print('Create recovery.conf file')
        recovery_config = template.render(
            name=self.name,
            host=self.master.host,
            port=self.master.port,
            version=self.version,
            username=self.master.username,
            password=self.master.password,
            base_destination=self.db_location,
            wal_destination=self.wal_location,
            slot=self.master.slot,
            delay=self.delay,
            dbname='postgres',
            standby_mode='true',
            standby_port=self.port,
        )

        file_location = f'/etc/postgresql/{self.version}/{self.name}/postgresql.conf' if self.version >= 12 \
            else f'{self.db_location}/recovery.conf'

        write_file(file_location, recovery_config)
        chown(file_location, 'postgres')

        if self.version >= 12:
            touch_file(f'{self.db_location}/recovery.signal')
            touch_file(f'{self.db_location}/standby.signal')

    def _get_current_delay(self):
        file_location = f'/etc/postgresql/{self.version}/{self.name}/postgresql.conf' if self.version >= 12 \
            else f'{self.db_location}/recovery.conf'

        config = read_file(file_location)
        for line in config.split('\n'):
            if 'recovery_min_apply_delay' in line:
                return int(line.split(' ')[2]) / 3600000

    def _get_replication_slot_status(self):
        return self.master.run_cmd('select active from pg_replication_slots')

    def _get_postgres_logs(self, lines):
        return execute_sync(f'tail -n {lines} /var/log/postgresql/postgresql-{self.version}-{self.name}.log')

    def _get_receive_wal_logs(self, lines):
        return execute_sync(f'tail -n {lines} /var/log/postgresql/receivewal-{self.version}-{self.name}.log')

    def print_status(self):
        print(f'# DELAY: {self._get_current_delay()}h')
        print()
        print('# REPLICATION SLOT:')
        print(self._get_replication_slot_status())
        print()
        print('# POSTGRES REPLICA SERVICE:')
        print(get_service_status(f'postgresql@{self.version}-{self.name}'))
        print()
        print('# RECEIVE WAL SERVICE:')
        print(get_service_status(f'receivewal-{self.version}-{self.name}'))
        print()
        print('# POSTGRES LOGS:')
        print(self._get_postgres_logs(10))
        print()
        print('# RECEIVE WAL LOGS:')
        print(self._get_receive_wal_logs(10))
